import { FormControl, FormGroup, FormArray, FormsModule, ReactiveFormsModule } from '@angular/forms';
import { map, startWith, distinctUntilChanged } from 'rxjs/operators';
import * as i0 from '@angular/core';
import { Injectable, Pipe, NgModule } from '@angular/core';
import { merge, of } from 'rxjs';
import { CommonModule } from '@angular/common';

function deepGet(object, path, throwOnMiss = false) {
    return path
        .reduce((partialState, pathSegment) => {
        if (throwOnMiss && typeof partialState[pathSegment] !== 'object') {
            throw new Error(`Path '${path}' could not be resolved in object ${object}. Segment '${pathSegment}' was not referring to type object.`);
        }
        return partialState[pathSegment] || {};
    }, object);
}
function deepEquals(x, y) {
    if (x === y) {
        return true; // if both x and y are null or undefined and exactly the same
    }
    else if (!(x instanceof Object) || !(y instanceof Object)) {
        return false; // if they are not strictly equal, they both need to be Objects
    }
    else if (x.constructor !== y.constructor) {
        // they must have the exact same prototype chain, the closest we can do is
        // test their constructor.
        return false;
    }
    else {
        for (const p in x) {
            if (!x.hasOwnProperty(p)) {
                continue; // other properties were tested using x.constructor === y.constructor
            }
            if (!y.hasOwnProperty(p)) {
                return false; // allows to compare x[ p ] and y[ p ] when set to undefined
            }
            if (x[p] === y[p]) {
                continue; // if they have the same strict value or identity then they are equal
            }
            if (typeof (x[p]) !== 'object') {
                return false; // Numbers, Strings, Functions, Booleans must be strictly equal
            }
            if (!deepEquals(x[p], y[p])) {
                return false;
            }
        }
        for (const p in y) {
            if (y.hasOwnProperty(p) && !x.hasOwnProperty(p)) {
                return false;
            }
        }
        return true;
    }
}
function isPrimitiveType(value) {
    return typeof value === 'string' || typeof value === 'number' || typeof value === 'boolean' || value === null;
}
function isPrimitiveListType(list) {
    return Array.isArray(list) && isPrimitiveType(list[0]);
}
function uuid(placeholder) {
    // @ts-ignore
    // tslint:disable-next-line
    return placeholder ? (placeholder ^ Math.random() * 16 >> placeholder / 4).toString(16) : ([1e7] + -1e3 + -4e3 + -8e3 + -1e11).replace(/[018]/g, uuid);
}
function normalizeKey(key) {
    if (typeof key === 'string') {
        return { id: key };
    }
    else {
        return key;
    }
}
function raiseError(message) {
    throw new Error(message);
}
function getAliasFromControl(control) {
    return control.formDefinition && control.formDefinition.alias && [control.formDefinition.alias] || [];
}
function extractErrorsWithAliasPrefix(control, prefixSeparator = '.') {
    if (!control.errors) {
        return null;
    }
    let path = getAliasFromControl(control);
    for (let parent = control.parentTypedControl; parent; parent = parent.parentTypedControl) {
        path = [...getAliasFromControl(parent), ...path];
    }
    return Object.keys(control.errors).reduce((prefixedErrors, key) => {
        const errors = control.errors || raiseError('Expected control to contain errors');
        prefixedErrors[[...path, key].join(prefixSeparator)] = errors[key];
        return prefixedErrors;
    }, {});
}

class TypedFormControl extends FormControl {
    constructor(formState, validatorOrOpts, asyncValidator, formDefinition) {
        super(formState, validatorOrOpts, asyncValidator);
        this.formDefinition = formDefinition;
    }
    get typedValue() {
        return this.value;
    }
    get typedValueChanges() {
        return this.valueChanges;
    }
    get errorsWithAliasPath() {
        return extractErrorsWithAliasPrefix(this);
    }
    get parentTypedControl() {
        return this.parent;
    }
    patchValue(value, options) {
        super.patchValue(value, options);
    }
    setValue(value, options) {
        super.setValue(value, options);
    }
    reset(formState, options) {
        super.reset(formState, options);
    }
    emitIfRequired(opts = {}) {
        if (opts.emitEvent !== false) {
            this.statusChanges.emit(this.status);
        }
    }
    markAsTouched(opts = {}) {
        super.markAsTouched({ onlySelf: opts.onlySelf });
        this.emitIfRequired(opts);
    }
    markAsUntouched(opts = {}) {
        super.markAsUntouched({ onlySelf: opts.onlySelf });
        this.emitIfRequired(opts);
    }
    markAsDirty(opts = {}) {
        super.markAsDirty({ onlySelf: opts.onlySelf });
        this.emitIfRequired(opts);
    }
    markAsPristine(opts = {}) {
        super.markAsPristine({ onlySelf: opts.onlySelf });
        this.emitIfRequired(opts);
    }
}
class TypedFormGroup extends FormGroup {
    constructor(controls, validatorOrOpts, asyncValidator, formDefinition) {
        super(controls, validatorOrOpts, asyncValidator);
        this.formDefinition = formDefinition;
    }
    get typedValue() {
        return this.value;
    }
    get typedRawValue() {
        return this.getRawValue();
    }
    get typedValueChanges() {
        return this.valueChanges;
    }
    get typedRawValueChanges() {
        return this.valueChanges.pipe(map(() => this.typedRawValue));
    }
    get typedControls() {
        return this.controls;
    }
    get errorsWithAliasPath() {
        return extractErrorsWithAliasPrefix(this);
    }
    get parentTypedControl() {
        return this.parent;
    }
    typedGet(key) {
        const childControl = super.get(key);
        return childControl;
    }
    typedGetCustomField(key) {
        const childControl = super.get(key);
        return childControl;
    }
    patchValue(value, options) {
        super.patchValue(value, options);
    }
    /** @ts-expect-error For the moment we need to omit override modifier because of parameter type incompatibility */
    setValue(value, options) {
        super.setValue(value, options);
    }
    reset(formState, options) {
        super.reset(formState, options);
    }
    emitIfRequired(opts = {}) {
        if (opts.emitEvent !== false) {
            this.statusChanges.emit(this.status);
        }
    }
    markAsTouched(opts = {}) {
        super.markAsTouched({ onlySelf: opts.onlySelf });
        this.emitIfRequired(opts);
    }
    markAsUntouched(opts = {}) {
        super.markAsUntouched({ onlySelf: opts.onlySelf });
        this.emitIfRequired(opts);
    }
    markAsDirty(opts = {}) {
        super.markAsDirty({ onlySelf: opts.onlySelf });
        this.emitIfRequired(opts);
    }
    markAsPristine(opts = {}) {
        super.markAsPristine({ onlySelf: opts.onlySelf });
        this.emitIfRequired(opts);
    }
}
class TypedFormArray extends FormArray {
    constructor(controls, validatorOrOpts, asyncValidator, formDefinition) {
        super(controls, validatorOrOpts, asyncValidator);
        this.formDefinition = formDefinition;
    }
    get typedValue() {
        return this.value;
    }
    get typedRawValue() {
        return this.getRawValue();
    }
    get typedControls() {
        return this.controls;
    }
    get typedValueChanges() {
        return this.valueChanges;
    }
    get typedRawValueChanges() {
        return this.valueChanges.pipe(map(() => this.typedRawValue));
    }
    get errorsWithAliasPath() {
        return extractErrorsWithAliasPrefix(this);
    }
    get parentTypedControl() {
        return this.parent;
    }
    typedAt(index) {
        const childControl = super.at(index);
        if (childControl instanceof TypedFormControl || childControl instanceof TypedFormGroup) {
            return childControl;
        }
        else {
            return null;
        }
    }
    typedAtCustomField(index) {
        return this.typedAt(index);
    }
    patchValue(value, options) {
        super.patchValue(value, options);
    }
    setValue(value, options) {
        super.setValue(value, options);
    }
    reset(formState, options) {
        super.reset(formState, options);
    }
    emitIfRequired(opts = {}) {
        if (opts.emitEvent !== false) {
            this.statusChanges.emit(this.status);
        }
    }
    markAsTouched(opts = {}) {
        super.markAsTouched({ onlySelf: opts.onlySelf });
        this.emitIfRequired(opts);
    }
    markAsUntouched(opts = {}) {
        super.markAsUntouched({ onlySelf: opts.onlySelf });
        this.emitIfRequired(opts);
    }
    markAsDirty(opts = {}) {
        super.markAsDirty({ onlySelf: opts.onlySelf });
        this.emitIfRequired(opts);
    }
    markAsPristine(opts = {}) {
        super.markAsPristine({ onlySelf: opts.onlySelf });
        this.emitIfRequired(opts);
    }
}

function createForm(formDefinitionOrInitialValue) {
    if (isPrimitiveType(formDefinitionOrInitialValue)) {
        return createForm({
            type: 'Field',
            initialValue: formDefinitionOrInitialValue
        });
    }
    else if (isPrimitiveListType(formDefinitionOrInitialValue)) {
        return createForm({
            type: 'PrimitiveArray',
            initialValue: formDefinitionOrInitialValue
        });
    }
    else if (formDefinitionOrInitialValue.type === 'Field' || formDefinitionOrInitialValue.type === 'CustomField') {
        return new TypedFormControl(formDefinitionOrInitialValue.initialValue, formDefinitionOrInitialValue.options, undefined, !isPrimitiveType(formDefinitionOrInitialValue) ? formDefinitionOrInitialValue : undefined);
    }
    else if (formDefinitionOrInitialValue.type === 'Group') {
        const controls = Object.keys(formDefinitionOrInitialValue.fields).reduce((fieldControls, fieldName) => {
            fieldControls[fieldName] = createForm(formDefinitionOrInitialValue.fields[fieldName]);
            return fieldControls;
        }, {});
        return new TypedFormGroup(controls, formDefinitionOrInitialValue.options, undefined, formDefinitionOrInitialValue);
    }
    else if (formDefinitionOrInitialValue.type === 'PrimitiveArray') {
        const controls = (formDefinitionOrInitialValue.initialValue || []).map(initialValue => createForm(initialValue));
        return new TypedFormArray(controls, formDefinitionOrInitialValue.options, undefined, !isPrimitiveListType(formDefinitionOrInitialValue) ? formDefinitionOrInitialValue : undefined);
    }
    else if (formDefinitionOrInitialValue.type === 'GroupArray') {
        const controls = Array.from({ length: formDefinitionOrInitialValue.initialItems || 0 }).map(() => createForm(formDefinitionOrInitialValue.group));
        return new TypedFormArray(controls, formDefinitionOrInitialValue.options, undefined, formDefinitionOrInitialValue);
    }
    else {
        throw new Error(`Invalid form definition or initial value ${formDefinitionOrInitialValue}`);
    }
}
function createNativeForm(formDefinitionOrInitialValue) {
    if (isPrimitiveType(formDefinitionOrInitialValue)) {
        return createNativeForm({
            type: 'Field',
            initialValue: formDefinitionOrInitialValue
        });
    }
    else if (isPrimitiveListType(formDefinitionOrInitialValue)) {
        return createNativeForm({
            type: 'PrimitiveArray',
            initialValue: formDefinitionOrInitialValue
        });
    }
    else if (formDefinitionOrInitialValue.type === 'Field' || formDefinitionOrInitialValue.type === 'CustomField') {
        return new FormControl(formDefinitionOrInitialValue.initialValue, formDefinitionOrInitialValue.options);
    }
    else if (formDefinitionOrInitialValue.type === 'Group') {
        const controls = Object.keys(formDefinitionOrInitialValue.fields).reduce((fieldControls, fieldName) => {
            fieldControls[fieldName] = createNativeForm(formDefinitionOrInitialValue.fields[fieldName]);
            return fieldControls;
        }, {});
        return new FormGroup(controls, formDefinitionOrInitialValue.options);
    }
    else if (formDefinitionOrInitialValue.type === 'PrimitiveArray') {
        const controls = (formDefinitionOrInitialValue.initialValue || []).map(initialValue => createNativeForm(initialValue));
        return new FormArray(controls, formDefinitionOrInitialValue.options);
    }
    else if (formDefinitionOrInitialValue.type === 'GroupArray') {
        const controls = Array.from({ length: formDefinitionOrInitialValue.initialItems || 0 }).map(() => createNativeForm(formDefinitionOrInitialValue.group));
        return new FormArray(controls, formDefinitionOrInitialValue.options, undefined);
    }
    else {
        throw new Error(`Invalid form definition or initial value ${formDefinitionOrInitialValue}`);
    }
}
function pushFormGroupArrayItem(formArrayDefinition, formArray, value, options) {
    if (formArrayDefinition.type === 'CustomField') {
        raiseError(`Cannot use pushFormGroupArrayItem on TypedFormArray with definition of type 'CustomField'`);
    }
    const formControl = createForm(formArrayDefinition.group);
    if (value) {
        formControl.setValue(value);
    }
    formArray.push(formControl, options);
    return formArray;
}

class FormRegistry {
    constructor() {
        this.forms = {};
    }
    createOrNormalizeFormKey(key) {
        return key ? normalizeKey(key) : { id: uuid() };
    }
    createAndRegisterForm(formDefinition, options) {
        let normalizedKey = null;
        if (options && options.key) {
            normalizedKey = normalizeKey(options.key);
        }
        if (normalizedKey && this.forms[normalizedKey.id]) {
            return normalizedKey;
        }
        else if (!normalizedKey) {
            normalizedKey = { id: uuid() };
        }
        this.forms[normalizedKey.id] = {
            form: createForm(formDefinition),
            formDefinition
        };
        if (options && options.initialData) {
            this.forms[normalizedKey.id].form.patchValue(options.initialData);
        }
        return normalizedKey;
    }
    registerForm(form, key) {
        const existingOrNewKey = this.createOrNormalizeFormKey(key);
        this.forms[existingOrNewKey.id] = {
            form
        };
        return existingOrNewKey;
    }
    getForm(key) {
        key = normalizeKey(key);
        if (!key || !this.forms[key.id]) {
            return null;
        }
        return this.forms[key.id].form;
    }
    removeForm(key) {
        key = normalizeKey(key);
        if (key && this.forms[key.id]) {
            delete this.forms[key.id];
            return true;
        }
        else {
            return false;
        }
    }
    clearForms() {
        this.forms = {};
    }
    containsForm(key) {
        key = normalizeKey(key);
        return !!this.forms[key.id];
    }
    observeForm(key) {
        const normalizedKey = normalizeKey(key);
        if (!normalizedKey || !this.forms[normalizedKey.id]) {
            return null;
        }
        const entry = this.forms[normalizedKey.id];
        return merge(entry.form.valueChanges.pipe(map(() => entry.form.value)), entry.form.statusChanges.pipe(map(() => entry.form.value))).pipe(startWith(entry.form.value), distinctUntilChanged((x, y) => deepEquals(x, y)));
    }
}
FormRegistry.ɵfac = i0.ɵɵngDeclareFactory({ minVersion: "12.0.0", version: "15.1.5", ngImport: i0, type: FormRegistry, deps: [], target: i0.ɵɵFactoryTarget.Injectable });
FormRegistry.ɵprov = i0.ɵɵngDeclareInjectable({ minVersion: "12.0.0", version: "15.1.5", ngImport: i0, type: FormRegistry, providedIn: 'root' });
i0.ɵɵngDeclareClassMetadata({ minVersion: "12.0.0", version: "15.1.5", ngImport: i0, type: FormRegistry, decorators: [{
            type: Injectable,
            args: [{
                    providedIn: 'root'
                }]
        }] });

class ObserveFormPipe {
    constructor(formRegistry, cdr) {
        this.formRegistry = formRegistry;
        this.cdr = cdr;
    }
    transform(formKey, { observe: observe = 'FormControl', strict: strict = true } = {}) {
        if (formKey === null || formKey === undefined) {
            if (strict) {
                throw new Error(`${formKey} was passed to ObserveFormPipe while in strict mode`);
            }
            return null;
        }
        const formData = this.formRegistry.observeForm(formKey);
        const formControl = this.formRegistry.getForm(formKey);
        if (!formData) {
            if (strict) {
                throw new Error(`ObserveFormPipe could not find form for key ${typeof formKey === 'string' ? formKey : formKey.id} while in strict mode`);
            }
            this.handleSubscription(of(null));
        }
        else if (observe === 'FormControl') {
            this.handleSubscription(formData.pipe(map(() => formControl)));
        }
        else {
            this.handleSubscription(formData);
        }
        this.latestReturnedValue = this.latestObservedValue;
        this.cdr.markForCheck();
        return this.latestObservedValue;
    }
    handleSubscription(observable) {
        if (this.latestObservable === observable) {
            return;
        }
        if (this.subscription) {
            this.subscription.unsubscribe();
        }
        this.subscription = observable.subscribe(value => {
            this.latestObservedValue = value;
            if (!deepEquals(this.latestObservedValue, this.latestReturnedValue)) {
                this.cdr.markForCheck();
            }
        });
    }
}
ObserveFormPipe.ɵfac = i0.ɵɵngDeclareFactory({ minVersion: "12.0.0", version: "15.1.5", ngImport: i0, type: ObserveFormPipe, deps: [{ token: FormRegistry }, { token: i0.ChangeDetectorRef }], target: i0.ɵɵFactoryTarget.Pipe });
ObserveFormPipe.ɵpipe = i0.ɵɵngDeclarePipe({ minVersion: "14.0.0", version: "15.1.5", ngImport: i0, type: ObserveFormPipe, name: "observeForm", pure: false });
i0.ɵɵngDeclareClassMetadata({ minVersion: "12.0.0", version: "15.1.5", ngImport: i0, type: ObserveFormPipe, decorators: [{
            type: Pipe,
            args: [{
                    name: 'observeForm',
                    pure: false
                }]
        }], ctorParameters: function () { return [{ type: FormRegistry }, { type: i0.ChangeDetectorRef }]; } });

class ReactiveFormsExtensionModule {
}
ReactiveFormsExtensionModule.ɵfac = i0.ɵɵngDeclareFactory({ minVersion: "12.0.0", version: "15.1.5", ngImport: i0, type: ReactiveFormsExtensionModule, deps: [], target: i0.ɵɵFactoryTarget.NgModule });
ReactiveFormsExtensionModule.ɵmod = i0.ɵɵngDeclareNgModule({ minVersion: "14.0.0", version: "15.1.5", ngImport: i0, type: ReactiveFormsExtensionModule, declarations: [ObserveFormPipe], imports: [CommonModule,
        FormsModule,
        ReactiveFormsModule], exports: [ObserveFormPipe] });
ReactiveFormsExtensionModule.ɵinj = i0.ɵɵngDeclareInjector({ minVersion: "12.0.0", version: "15.1.5", ngImport: i0, type: ReactiveFormsExtensionModule, imports: [CommonModule,
        FormsModule,
        ReactiveFormsModule] });
i0.ɵɵngDeclareClassMetadata({ minVersion: "12.0.0", version: "15.1.5", ngImport: i0, type: ReactiveFormsExtensionModule, decorators: [{
            type: NgModule,
            args: [{
                    imports: [
                        CommonModule,
                        FormsModule,
                        ReactiveFormsModule
                    ],
                    declarations: [
                        ObserveFormPipe
                    ],
                    exports: [
                        ObserveFormPipe
                    ]
                }]
        }] });

/**
 * Generated bundle index. Do not edit.
 */

export { FormRegistry, ObserveFormPipe, ReactiveFormsExtensionModule, TypedFormArray, TypedFormControl, TypedFormGroup, createForm, pushFormGroupArrayItem };
//# sourceMappingURL=ng-rfx.mjs.map
